from pathlib import Path

import duckdb
import pytest
from pandas.testing import assert_frame_equal

from .utils import to_pandas  # noqa: TID252


@pytest.fixture(scope="module")
def duck():
    conn = duckdb.connect()
    conn.sql("CALL dbgen(sf = 0.001)")
    return conn


@pytest.fixture(scope="module", autouse=True)
def data(sail, spark, duck):  # noqa: ARG001
    tables = list(duck.sql("SHOW TABLES").df()["name"])
    for table in tables:
        df = duck.sql(f"SELECT * FROM {table}").arrow().to_pandas()  # noqa: S608
        sail.createDataFrame(df).createOrReplaceTempView(table)
        # spark.createDataFrame(df).createOrReplaceTempView(table)
    yield
    for table in tables:
        sail.catalog.dropTempView(table)
        # spark.catalog.dropTempView(table)


@pytest.mark.parametrize("query", [f"q{x + 1}" for x in range(22)])
def test_derived_tpch_query_execution(sail, query):
    for sql in read_sql(query):
        sail.sql(sql).toPandas()


@pytest.mark.parametrize("query", [f"q{x + 1}" for x in range(22)])
@pytest.mark.skip(reason="Spark data loading is not reliable")
def test_derived_tpch_query_spark_parity(sail, spark, query):
    for sql in read_sql(query):
        actual = sail.sql(sql)
        expected = spark.sql(sql)
        if is_ddl(sql):
            continue
        # TODO: improve data type parity with Spark
        assert_frame_equal(to_pandas(actual), to_pandas(expected), check_exact=False, check_dtype=False, atol=1e-3)


def read_sql(query):
    path = Path(__file__).parent.parent.parent / "data" / "tpch" / "queries" / f"{query}.sql"
    with open(path) as f:
        text = f.read()
    for sql in text.split(";"):
        sql = sql.strip()  # noqa: PLW2901
        sql = sql.replace("create view", "create temp view")  # noqa: PLW2901
        if sql:
            yield sql


def is_ddl(sql):
    return any(x in sql for x in ("create view", "create temp view", "drop view"))



import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from pyspark.sql.types import Row


@pytest.mark.parametrize(
    ("sql", "expected"),
    [
        ("SELECT * FROM range(-1)", []),
        ("SELECT * FROM range(1)", [0]),
        ("SELECT * FROM range(2, 5)", [2, 3, 4]),
        ("SELECT * FROM range(1 + 2)", [0, 1, 2]),
        ("SELECT * FROM range(CAST('2' AS INT))", [0, 1]),
        ("SELECT * FROM range(3, 0, -1)", [1, 2, 3]),
        ("SELECT * FROM range(10, 0, -2, 3)", [2, 4, 6, 8, 10]),
    ],
)
def test_range(sail, sql, expected):
    assert_frame_equal(
        sail.sql(sql).toPandas().sort_values("id").reset_index(drop=True),
        pd.DataFrame({"id": expected}, dtype="int64"),
    )


def test_lateral_view(sail):
    df = sail.sql("""
        SELECT * FROM range(2)
            LATERAL VIEW explode(array(id, id + 1)) AS v
    """)
    assert_frame_equal(
        df.toPandas(),
        pd.DataFrame({"id": [0, 0, 1, 1], "v": [0, 1, 1, 2]}, dtype="int64"),
    )

    df = sail.sql("""
        SELECT * FROM range(2)
            LATERAL VIEW explode(array(id, id + 1)) t AS u
            LATERAL VIEW explode(array(u, t.u * 2)) AS v
    """)
    assert_frame_equal(
        df.toPandas(),
        pd.DataFrame(
            {"id": [0, 0, 0, 0, 1, 1, 1, 1], "u": [0, 0, 1, 1, 1, 1, 2, 2], "v": [0, 0, 1, 2, 1, 2, 2, 4]},
            dtype="int64",
        ),
    )


def test_lateral_view_outer(sail):
    df = sail.sql("""
        SELECT * FROM range(1)
            LATERAL VIEW explode(CAST(NULL AS array<int>)) AS v
    """)
    assert df.collect() == []

    df = sail.sql("""
        SELECT * FROM range(1)
            LATERAL VIEW OUTER explode(CAST(NULL AS array<int>)) AS v
    """)
    assert df.collect() == [Row(id=0, v=None)]

    df = sail.sql("""
        SELECT * FROM range(1)
            LATERAL VIEW explode_outer(CAST(NULL AS array<int>)) AS v
    """)
    assert df.collect() == [Row(id=0, v=None)]


def test_lateral_join(sail):
    df = sail.sql("""
        SELECT * FROM range(1), LATERAL explode(array(id, id + 1)) AS t(v)
    """)
    assert_frame_equal(
        df.toPandas(),
        pd.DataFrame({"id": [0, 0], "v": [0, 1]}, dtype="int64"),
    )

    df = sail.sql("""
        SELECT * FROM range(1) JOIN LATERAL explode(array(id, id + 1)) AS t(v)
    """)
    assert_frame_equal(
        df.toPandas(),
        pd.DataFrame({"id": [0, 0], "v": [0, 1]}, dtype="int64"),
    )


def test_lateral_join_without_table(sail):
    df = sail.sql("""
        SELECT * FROM LATERAL explode(array(0, 1)) AS t(v)
    """)
    assert_frame_equal(
        df.toPandas(),
        pd.DataFrame({"v": [0, 1]}, dtype="int32"),
    )



import datetime
from decimal import Decimal

import pytest
from pyspark.sql.types import Row


def test_plus(sail):
    assert sail.sql("SELECT 1 + 2 AS result;").collect() == [Row(result=3)]
    assert sail.sql("SELECT DATE'2025-02-26' + INTERVAL '2' MONTH AS result;").collect() == [
        Row(result=datetime.date(2025, 4, 26))
    ]
    assert sail.sql("SELECT TIMESTAMP'2025-02-26 12:15:29' + INTERVAL '3' SECOND AS result;").collect() == [
        Row(result=datetime.datetime(2025, 2, 26, 12, 15, 32))  # noqa: DTZ001
    ]
    assert sail.sql("SELECT DATE'2025-03-31' + INTERVAL '1' MONTH AS result;").collect() == [
        Row(result=datetime.date(2025, 4, 30))
    ]
    assert sail.sql("SELECT DATE'2025-02-26' + 1 AS result;").collect() == [Row(result=datetime.date(2025, 2, 27))]
    assert sail.sql("SELECT 1 + DATE'2025-02-26' AS result;").collect() == [Row(result=datetime.date(2025, 2, 27))]


def test_minus(sail):
    assert sail.sql("SELECT 2 - 1 AS result;").collect() == [Row(result=1)]
    assert sail.sql("SELECT TIMESTAMP'2025-02-26 12:15:29' - INTERVAL '3' SECOND AS result;").collect() == [
        Row(result=datetime.datetime(2025, 2, 26, 12, 15, 26))  # noqa: DTZ001
    ]
    assert sail.sql("SELECT DATE'2025-03-31' - INTERVAL '1' MONTH AS result;").collect() == [
        Row(result=datetime.date(2025, 2, 28))
    ]
    assert sail.sql("SELECT DATE'2025-02-26' - 1 AS result;").collect() == [Row(result=datetime.date(2025, 2, 25))]
    with pytest.raises(Exception, match="Cannot coerce arithmetic expression Int32 - Date32 to valid types"):
        sail.sql("SELECT 1 - DATE'2025-02-26' AS result;").collect()


def test_multiply(sail):
    assert sail.sql("SELECT 3 * 2 AS result;").collect() == [Row(result=6)]
    assert sail.sql("SELECT 2L * 2L AS result;").collect() == [Row(result=4)]
    # Not supported in Sail yet:
    # assert sail.sql("SELECT INTERVAL '3:15' HOUR TO MINUTE * 3 AS result;").collect() == [
    #     Row(result=datetime.timedelta(seconds=35100))
    # ]
    # assert sail.sql("SELECT INTERVAL '3' YEAR * 3 AS result;").collect() == [Row(result=108)]


def test_divide(sail):
    assert sail.sql("SELECT 3 / 2 AS result;").collect() == [Row(result=1.5)]
    assert sail.sql("SELECT 2L / 2L AS result;").collect() == [Row(result=1.0)]
    assert sail.sql("SELECT 6 / Decimal(3) AS result;").collect() == [Row(result=Decimal("2.00000000000"))]
    # TODO: Spark outputs Decimal("2.000000") but Sail outputs Decimal("2.00000000000")
    assert sail.sql("SELECT Decimal(6) / 3 AS result;").collect() == [Row(result=Decimal("2.00000000000"))]
    # TODO: Spark outputs Decimal("1.50000000000") but Sail outputs Decimal("1.50000")
    assert sail.sql("SELECT Decimal(3) / Decimal(2) AS result;").collect() == [Row(result=Decimal("1.50000"))]
    # TODO: Spark outputs Decimal("1.500000") but Sail outputs Decimal("1.50000")
    assert sail.sql("SELECT Decimal(3) / 2 AS result;").collect() == [Row(result=Decimal("1.50000"))]
    # TODO: Spark outputs Decimal("0.66666666667") but Sail outputs Decimal("0.6666")
    assert sail.sql("SELECT 2 / Decimal(3) AS result;").collect() == [Row(result=Decimal("0.6666"))]
    # TODO: Spark outputs Decimal("0.666667") but Sail outputs Decimal("0.6666")
    assert sail.sql("SELECT 2.0 / 3.0 AS result;").collect() == [Row(result=Decimal("0.66666"))]
    # Not supported in Sail yet:
    # assert sail.sql("SELECT INTERVAL '3:15' HOUR TO MINUTE / 3 AS result;").collect() == [
    #     Row(result=datetime.timedelta(seconds=3900))
    # ]


def test_div(sail):
    assert sail.sql("SELECT 3 div 2 AS result;").collect() == [Row(result=1)]
    assert sail.sql("SELECT -5.9 div 1;").collect() == [Row(result=-5)]
    # Not supported in Sail yet:
    # assert sail.sql("SELECT INTERVAL '10' DAY div INTERVAL '3' DAY AS result;").collect() == [Row(result=3)]
    # assert sail.sql("SELECT INTERVAL '1-1' YEAR TO MONTH div INTERVAL '-1' MONTH AS result;").collect() == [Row(result=-13)]
    # assert sail.sql("SELECT INTERVAL '100' HOUR div INTERVAL '1' DAY AS result;").collect() == [Row(result=4)]



import random

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal


@pytest.fixture(scope="module", autouse=True)
def person_table(sail):
    sail.sql("CREATE TABLE person (id INT, name STRING DEFAULT 'Sail', age INT)")
    yield
    sail.sql("DROP TABLE person")


def test_insert_single_value(sail):
    p_id = random.randint(0, 1000000)  # noqa: S311
    sail.sql(f"INSERT INTO person VALUES ({p_id}, 'Shehab', 99)")  # noqa: S608
    actual = sail.sql(f"SELECT * FROM person where id = {p_id}").toPandas()  # noqa: S608
    expected = pd.DataFrame(
        {"id": [p_id], "name": ["Shehab"], "age": [99]},
    ).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)


def test_insert_multiple_values(sail):
    p_id1 = random.randint(0, 1000000)  # noqa: S311
    p_id2 = random.randint(0, 1000000)  # noqa: S311
    p_id3 = random.randint(0, 1000000)  # noqa: S311
    sail.sql(f"INSERT INTO person VALUES ({p_id1}, 'Shehab', 99), ({p_id2}, 'Heran', 2), ({p_id3}, 'Chen', 9000)")  # noqa: S608
    actual = sail.sql(f"SELECT * FROM person WHERE id IN ({p_id1}, {p_id2}, {p_id3})").toPandas()  # noqa: S608
    expected = pd.DataFrame(
        {"id": [p_id1, p_id2, p_id3], "name": ["Shehab", "Heran", "Chen"], "age": [99, 2, 9000]},
    ).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)


def test_insert_using_column_list(sail):
    p_id1 = random.randint(0, 1000000)  # noqa: S311
    p_id2 = random.randint(0, 1000000)  # noqa: S311
    p_id3 = random.randint(0, 1000000)  # noqa: S311
    sail.sql(
        f"INSERT INTO person (id, name, age) VALUES ({p_id1}, 'Shehab', 99), ({p_id2}, 'Heran', 2), ({p_id3}, 'Chen', 9000)"  # noqa: S608
    )
    actual = sail.sql(f"SELECT * FROM person WHERE id IN ({p_id1}, {p_id2}, {p_id3})").toPandas()  # noqa: S608
    expected = pd.DataFrame(
        {"id": [p_id1, p_id2, p_id3], "name": ["Shehab", "Heran", "Chen"], "age": [99, 2, 9000]},
    ).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)

    p_id1 = random.randint(0, 1000000)  # noqa: S311
    p_id2 = random.randint(0, 1000000)  # noqa: S311
    p_id3 = random.randint(0, 1000000)  # noqa: S311
    sail.sql(
        f"INSERT INTO person (name, age, id) VALUES ('Shehabz', 99, {p_id1}), ('Heranz', 2, {p_id2}), ('Chenz', 9000, {p_id3})"  # noqa: S608
    )
    actual = sail.sql(f"SELECT * FROM person WHERE id IN ({p_id1}, {p_id2}, {p_id3})").toPandas()  # noqa: S608
    expected = pd.DataFrame(
        {"id": [p_id1, p_id2, p_id3], "name": ["Shehabz", "Heranz", "Chenz"], "age": [99, 2, 9000]},
    ).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)


def test_insert_using_implicit_default(sail):
    p_id1 = random.randint(0, 1000000)  # noqa: S311
    p_id2 = random.randint(0, 1000000)  # noqa: S311
    p_id3 = random.randint(0, 1000000)  # noqa: S311
    sail.sql(f"INSERT INTO person (id, age) VALUES ({p_id1}, 99), ({p_id2}, 2), ({p_id3}, 9000)")  # noqa: S608
    actual = sail.sql(f"SELECT * FROM person WHERE id IN ({p_id1}, {p_id2}, {p_id3})").toPandas()  # noqa: S608
    expected = pd.DataFrame(
        {"id": [p_id1, p_id2, p_id3], "name": ["Sail", "Sail", "Sail"], "age": [99, 2, 9000]},
    ).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)

    p_id1 = random.randint(0, 1000000)  # noqa: S311
    p_id2 = random.randint(0, 1000000)  # noqa: S311
    p_id3 = random.randint(0, 1000000)  # noqa: S311
    sail.sql(f"INSERT INTO person (age, id) VALUES (100, {p_id1}), (3, {p_id2}), (9001, {p_id3})")  # noqa: S608
    actual = sail.sql(f"SELECT * FROM person WHERE id IN ({p_id1}, {p_id2}, {p_id3})").toPandas()  # noqa: S608
    expected = pd.DataFrame(
        {"id": [p_id1, p_id2, p_id3], "name": ["Sail", "Sail", "Sail"], "age": [100, 3, 9001]},
    ).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)


def test_insert_with_full_table_name(sail):
    p_id = random.randint(0, 1000000)  # noqa: S311
    sail.sql(f"INSERT INTO spark_catalog.default.person VALUES ({p_id}, 'Shehab', 99)")  # noqa: S608
    actual = sail.sql(f"SELECT * FROM person WHERE id = {p_id}").toPandas()  # noqa: S608
    expected = pd.DataFrame(
        {"id": [p_id], "name": ["Shehab"], "age": [99]},
    ).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)


@pytest.mark.skip(reason="DELETE ExecutionPlan not implemented")
def test_delete_entire_table(sail):
    sail.sql("CREATE TABLE meow (id INT, name STRING DEFAULT 'Sail', age INT)")
    p_id1 = random.randint(0, 1000000)  # noqa: S311
    p_id2 = random.randint(0, 1000000)  # noqa: S311
    p_id3 = random.randint(0, 1000000)  # noqa: S311
    sail.sql(f"INSERT INTO meow VALUES ({p_id1}, 'Shehab', 99), ({p_id2}, 'Heran', 2), ({p_id3}, 'Chen', 9000)")  # noqa: S608
    sail.sql("DELETE FROM meow")
    actual = sail.sql("SELECT * FROM meow").toPandas()
    expected = pd.DataFrame(columns=["id", "name", "age"]).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)


@pytest.mark.skip(reason="DELETE ExecutionPlan not implemented")
def test_delete_with_filter(sail):
    # TODO: Add test for DELETE with table alias and filter
    p_id = random.randint(0, 1000000)  # noqa: S311
    sail.sql(f"INSERT INTO person VALUES ({p_id}, 'Shehab', 99)")  # noqa: S608
    sail.sql(f"DELETE FROM person WHERE id = {p_id}")  # noqa: S608
    actual = sail.sql(f"SELECT * FROM person WHERE id = {p_id}").toPandas()  # noqa: S608
    expected = pd.DataFrame(columns=["id", "name", "age"]).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)


@pytest.mark.skip(reason="UPDATE ExecutionPlan not implemented")
def test_update_single_value(sail):
    p_id = random.randint(0, 1000000)  # noqa: S311
    p_id1 = p_id + 1
    sail.sql(f"INSERT INTO person VALUES ({p_id}, 'Shehab', 99)")  # noqa: S608
    sail.sql(f"UPDATE person SET id = {p_id1}, age = 100 WHERE id = {p_id}")  # noqa: S608
    actual = sail.sql(f"SELECT * FROM person WHERE id = {p_id1}").toPandas()  # noqa: S608
    expected = pd.DataFrame(
        {"id": [p_id1], "name": ["Shehab"], "age": [100]},
    ).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)


@pytest.mark.skip(reason="UPDATE ExecutionPlan not implemented")
def test_update_multiple_values(sail):
    p_id1 = random.randint(0, 1000000)  # noqa: S311
    p_id2 = random.randint(0, 1000000)  # noqa: S311
    p_id3 = random.randint(0, 1000000)  # noqa: S311
    sail.sql(f"INSERT INTO person VALUES ({p_id1}, 'Shehab', 99), ({p_id2}, 'Heran', 2), ({p_id3}, 'Chen', 9000)")  # noqa: S608
    sail.sql(f"UPDATE person SET id = {p_id1+1}, age = 100 WHERE id IN ({p_id1}, {p_id2}, {p_id3})")  # noqa: S608
    actual = sail.sql(f"SELECT * FROM person WHERE id IN ({p_id1+1}, {p_id2+1}, {p_id3+1})").toPandas()  # noqa: S608
    expected = pd.DataFrame(
        {"id": [p_id1 + 1, p_id1 + 1, p_id1 + 1], "name": ["Shehab", "Heran", "Chen"], "age": [100, 100, 100]},
    ).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)


@pytest.mark.skip(reason="UPDATE ExecutionPlan not implemented")
def test_update_with_full_table_name(sail):
    p_id = random.randint(0, 1000000)  # noqa: S311
    sail.sql(f"INSERT INTO person VALUES ({p_id}, 'Shehab', 99)")  # noqa: S608
    sail.sql(f"UPDATE spark_catalog.default.person SET id = {p_id+1}, age = 100 WHERE id = {p_id}")  # noqa: S608
    actual = sail.sql(f"SELECT * FROM person WHERE id = {p_id+1}").toPandas()  # noqa: S608
    expected = pd.DataFrame(
        {"id": [p_id + 1], "name": ["Shehab"], "age": [100]},
    ).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)


@pytest.mark.skip(reason="UPDATE ExecutionPlan not implemented")
def test_update_with_alias(sail):
    p_id = random.randint(0, 1000000)  # noqa: S311
    sail.sql(f"INSERT INTO person VALUES ({p_id}, 'Shehab', 99)")  # noqa: S608
    sail.sql(f"UPDATE person AS p SET p.id = {p_id+1}, p.age = 100 WHERE p.id = {p_id}")  # noqa: S608
    actual = sail.sql(f"SELECT * FROM person WHERE id = {p_id+1}").toPandas()  # noqa: S608
    expected = pd.DataFrame(
        {"id": [p_id + 1], "name": ["Shehab"], "age": [100]},
    ).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)


@pytest.mark.skip(reason="UPDATE ExecutionPlan not implemented")
def test_update_with_full_table_name_and_alias(sail):
    p_id = random.randint(0, 1000000)  # noqa: S311
    sail.sql(f"INSERT INTO person VALUES ({p_id}, 'Shehab', 99)")  # noqa: S608
    sail.sql(f"UPDATE spark_catalog.default.person AS p SET p.id = {p_id+1}, p.age = 100 WHERE p.id = {p_id}")  # noqa: S608
    actual = sail.sql(f"SELECT * FROM person WHERE id = {p_id+1}").toPandas()  # noqa: S608
    expected = pd.DataFrame(
        {"id": [p_id + 1], "name": ["Shehab"], "age": [100]},
    ).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)


@pytest.mark.skip(reason="UPDATE ExecutionPlan not implemented")
def test_update_using_column_value(sail):
    p_id = random.randint(0, 1000000)  # noqa: S311
    p_id1 = p_id + 1
    sail.sql(f"INSERT INTO person VALUES ({p_id}, 'Shehab', 99)")  # noqa: S608
    sail.sql(f"UPDATE person SET id = id+1, age = 100 WHERE id = {p_id}")  # noqa: S608
    actual = sail.sql(f"SELECT * FROM person WHERE id = {p_id1}").toPandas()  # noqa: S608
    expected = pd.DataFrame(
        {"id": [p_id1], "name": ["Shehab"], "age": [100]},
    ).astype({"id": "int32", "name": "str", "age": "int32"})
    assert_frame_equal(actual, expected)



import os

import pytest
from pyspark.context import SparkContext
from pyspark.sql import SparkSession

from pysail.spark import SparkConnectServer


@pytest.fixture(scope="session")
def remote():
    if r := os.environ.get("SPARK_REMOTE"):
        yield r
    else:
        server = SparkConnectServer("127.0.0.1", 0)
        server.start(background=True)
        _, port = server.listening_address
        yield f"sc://localhost:{port}"
        server.stop()


@pytest.fixture(scope="session")
def sail(remote):
    if "SPARK_LOCAL_REMOTE" in os.environ:
        del os.environ["SPARK_LOCAL_REMOTE"]
    sail = SparkSession.builder.remote(remote).appName("Sail").getOrCreate()
    configure_spark_session(sail)
    yield sail
    sail.stop()


@pytest.fixture(scope="session")
def spark():
    os.environ["SPARK_LOCAL_IP"] = "127.0.0.1"
    os.environ["SPARK_LOCAL_REMOTE"] = "true"
    sc = SparkContext("local", "Spark")
    spark = SparkSession(sc)
    configure_spark_session(spark)
    yield spark
    spark.stop()
    if "SPARK_LOCAL_REMOTE" in os.environ:
        del os.environ["SPARK_LOCAL_REMOTE"]


def configure_spark_session(session: SparkSession):
    # Set the Spark session time zone to UTC by default.
    # Some test data (e.g. TPC-DS data) may generate timestamps that is invalid
    # in some local time zones. This would result in `pytz.exceptions.NonExistentTimeError`
    # when converting such timestamps from the local time zone to UTC.
    session.conf.set("spark.sql.session.timeZone", "UTC")
    # Enable Arrow to avoid data type errors when creating Spark DataFrame from Pandas.
    session.conf.set("spark.sql.execution.arrow.pyspark.enabled", "true")


@pytest.fixture(scope="session", autouse=True)
def sail_doctest(doctest_namespace, sail):
    # TODO: we may need to isolate the Spark session for each doctest
    #   so that the registered temporary views and UDFs does not interfere
    #   with each other.
    doctest_namespace["spark"] = sail



import sys

from pysail import _native


def main():
    # When the Sail CLI is invoked via `python -m pysail`, the first argument in `sys.argv` is
    # the absolute path to `__main__.py`.
    _native.main(sys.argv)



import pandas as pd
from pyspark.sql.types import DecimalType, DoubleType, FloatType


def to_pandas(df):
    """Converts a Spark DataFrame to a Pandas DataFrame.
    This function additionally converts columns of floating-point types to the
    Pandas nullable float64 type. Otherwise, such columns may have the
    `object` type in Pandas, and approximate comparison would not work.
    """

    def _to_pandas_type(dt):
        if isinstance(dt, (FloatType, DoubleType, DecimalType)):
            return pd.Float64Dtype()
        return None

    dtypes = {f.name: dt for f in df.schema.fields if (dt := _to_pandas_type(f.dataType)) is not None}
    return df.toPandas().astype(dtypes)



import pytest
from pyspark.sql.types import Row


def test_get_item_ignore_case(sail):
    df = sail.sql("SELECT struct(1 AS b) AS a")
    assert df.select(df.a.getItem("b")).collect() == [Row(**{"a.b": 1})]
    assert df.select(df.a.getItem("B")).collect() == [Row(**{"a.B": 1})]


@pytest.mark.skip(reason="not working")
def test_get_item_nested_map(sail):
    df = sail.sql("SELECT struct(map(1, 2) AS b) AS a")
    assert df.select(df.a.getItem("b").getItem(1)).collect() == [Row(**{"a.b[1]": 2})]
    df = sail.sql("SELECT map('b', map(1, 2)) AS a")
    assert df.select(df.a.getItem("b").getItem(1)).collect() == [Row(**{"a[b][1]": 2})]



def pytest_configure(config):
    """Configure pytest.

    We include the tests in the installed package so that the user can test the installation
    via `pytest --pyargs pysail`.
    We must customize the configuration here instead of using `pytest.ini` or `pyproject.toml`
    since these files are not part of the installed package.
    """

    # Note: configuration set via `config.inicfg` may not have an effect due to the cache used
    # in `config.getini()`. In such a case, we may have to clear the INI cache in `config`.
    # Since clearing the cache requires access to the private attribute of `config`, we do not
    # do it here unless absolutely necessary, to avoid compatibility issues with future versions
    # of pytest.
    # In common cases, plugins only access the configuration for the first time after this hook,
    # so the cache is not a problem.

    config.inicfg["doctest_optionflags"] = "ELLIPSIS NORMALIZE_WHITESPACE IGNORE_EXCEPTION_DETAIL NUMBER"



import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

# The test cases are adapted from the Spark SQL documentation.
# https://spark.apache.org/docs/latest/sql-ref-syntax-qry-select-groupby.html


@pytest.fixture(scope="module", autouse=True)
def dealer_table(sail):
    df = sail.createDataFrame(
        [
            (100, "Fremont", "Honda Civic", 10),
            (100, "Fremont", "Honda Accord", 15),
            (100, "Fremont", "Honda CRV", 7),
            (200, "Dublin", "Honda Civic", 20),
            (200, "Dublin", "Honda Accord", 10),
            (200, "Dublin", "Honda CRV", 3),
            (300, "San Jose", "Honda Civic", 5),
            (300, "San Jose", "Honda Accord", 8),
        ],
        schema="id INT, city STRING, car_model STRING, quantity INT",
    )
    name = "dealer"
    df.createOrReplaceTempView(name)
    yield
    sail.catalog.dropTempView(name)


@pytest.fixture(scope="module", autouse=True)
def person_table(sail):
    df = sail.createDataFrame(
        [(100, "Mary", None), (200, "John", 30), (300, "Mike", 80), (400, "Dan", 50)],
        schema="id INT, name STRING, age INT",
    )
    name = "person"
    df.createOrReplaceTempView(name)
    yield
    sail.catalog.dropTempView(name)


def test_group_by(sail):
    actual = sail.sql("SELECT id, sum(quantity) FROM dealer GROUP BY id ORDER BY id").toPandas()
    expected = pd.DataFrame(
        {"id": [100, 200, 300], "sum(quantity)": [32, 33, 13]},
    ).astype({"id": "int32", "sum(quantity)": "int64"})
    assert_frame_equal(actual, expected)


def test_group_by_column_position(sail):
    actual = sail.sql("SELECT id, sum(quantity) FROM dealer GROUP BY 1 ORDER BY 1").toPandas()
    expected = pd.DataFrame(
        {"id": [100, 200, 300], "sum(quantity)": [32, 33, 13]},
    ).astype({"id": "int32", "sum(quantity)": "int64"})
    assert_frame_equal(actual, expected)


def test_multiple_aggregations(sail):
    actual = sail.sql("""
        SELECT id, sum(quantity) AS sum, max(quantity) AS max
        FROM dealer
        GROUP BY id
        ORDER BY id
    """).toPandas()
    expected = pd.DataFrame({"id": [100, 200, 300], "sum": [32, 33, 13], "max": [15, 20, 8]}).astype(
        {"id": "int32", "sum": "int64", "max": "int32"}
    )
    assert_frame_equal(actual, expected)


def test_count_distinct(sail):
    actual = sail.sql("""
        SELECT car_model, count(DISTINCT city) AS count
        FROM dealer
        GROUP BY car_model
    """).toPandas()
    expected = pd.DataFrame({"car_model": ["Honda Civic", "Honda CRV", "Honda Accord"], "count": [3, 2, 3]})

    def sort(df):
        return df.sort_values("car_model", ignore_index=True)

    assert_frame_equal(sort(actual), sort(expected))


@pytest.mark.skip(reason="not implemented")
def test_aggregation_filter(sail):
    actual = sail.sql("""
        SELECT id, sum(quantity) FILTER (
            WHERE car_model IN ('Honda Civic', 'Honda CRV')
        ) AS `sum(quantity)`
        FROM dealer
        GROUP BY id
        ORDER BY id
    """).toPandas()
    expected = pd.DataFrame({"id": [100, 200, 300], "sum(quantity)": [17, 23, 5]}).astype(
        {"id": "int32", "sum(quantity)": "int64"}
    )
    assert_frame_equal(actual, expected)


def test_grouping_sets(sail):
    actual = sail.sql("""
        SELECT city, car_model, sum(quantity) AS sum
        FROM dealer
        GROUP BY GROUPING SETS ((city, car_model), (city), (car_model), ())
        ORDER BY city, car_model
    """).toPandas()
    expected = pd.DataFrame(
        {
            "city": [
                None,
                None,
                None,
                None,
                "Dublin",
                "Dublin",
                "Dublin",
                "Dublin",
                "Fremont",
                "Fremont",
                "Fremont",
                "Fremont",
                "San Jose",
                "San Jose",
                "San Jose",
            ],
            "car_model": [
                None,
                "Honda Accord",
                "Honda CRV",
                "Honda Civic",
                None,
                "Honda Accord",
                "Honda CRV",
                "Honda Civic",
                None,
                "Honda Accord",
                "Honda CRV",
                "Honda Civic",
                None,
                "Honda Accord",
                "Honda Civic",
            ],
            "sum": [78, 33, 10, 35, 33, 10, 3, 20, 32, 15, 7, 10, 13, 8, 5],
        }
    )
    assert_frame_equal(actual, expected)


def test_rollup(sail):
    actual = sail.sql("""
        SELECT city, car_model, sum(quantity) AS sum
        FROM dealer
        GROUP BY city, car_model WITH ROLLUP
        ORDER BY city, car_model
    """).toPandas()
    expected = pd.DataFrame(
        {
            "city": [
                None,
                "Dublin",
                "Dublin",
                "Dublin",
                "Dublin",
                "Fremont",
                "Fremont",
                "Fremont",
                "Fremont",
                "San Jose",
                "San Jose",
                "San Jose",
            ],
            "car_model": [
                None,
                None,
                "Honda Accord",
                "Honda CRV",
                "Honda Civic",
                None,
                "Honda Accord",
                "Honda CRV",
                "Honda Civic",
                None,
                "Honda Accord",
                "Honda Civic",
            ],
            "sum": [78, 33, 10, 3, 20, 32, 15, 7, 10, 13, 8, 5],
        }
    )
    assert_frame_equal(actual, expected)


def test_cube(sail):
    actual = sail.sql("""
        SELECT city, car_model, sum(quantity) AS sum
        FROM dealer
        GROUP BY city, car_model WITH CUBE
        ORDER BY city, car_model
    """).toPandas()
    expected = pd.DataFrame(
        {
            "city": [
                None,
                None,
                None,
                None,
                "Dublin",
                "Dublin",
                "Dublin",
                "Dublin",
                "Fremont",
                "Fremont",
                "Fremont",
                "Fremont",
                "San Jose",
                "San Jose",
                "San Jose",
            ],
            "car_model": [
                None,
                "Honda Accord",
                "Honda CRV",
                "Honda Civic",
                None,
                "Honda Accord",
                "Honda CRV",
                "Honda Civic",
                None,
                "Honda Accord",
                "Honda CRV",
                "Honda Civic",
                None,
                "Honda Accord",
                "Honda Civic",
            ],
            "sum": [78, 33, 10, 35, 33, 10, 3, 20, 32, 15, 7, 10, 13, 8, 5],
        }
    )
    assert_frame_equal(actual, expected)


def test_aggregation_with_nulls(sail):
    actual = sail.sql("SELECT FIRST(age) FROM person").toPandas()
    expected = pd.DataFrame({"first(age)": [None]}).astype({"first(age)": "float64"})
    assert_frame_equal(actual, expected)


@pytest.mark.skip(reason="not implemented")
def test_aggregation_ignore_nulls(sail):
    actual = sail.sql("SELECT FIRST(age IGNORE NULLS), LAST(id), SUM(id) FROM person").toPandas()
    expected = pd.DataFrame({"first(age)": [30], "last(id)": [400], "sum(id)": [1000]})
    assert_frame_equal(actual, expected)



import pandas as pd
from pandas.testing import assert_frame_equal


def test_map_in_pandas(sail):
    df = sail.createDataFrame([(1, "Alice"), (2, "Bob")], schema="id long, name string")

    def f(it):
        for pdf in it:
            pdf["id"] = pdf["id"] + 1
            pdf["name"] = pdf["name"].str.upper()
            yield pdf

    actual = df.mapInPandas(f, "id long, name string").toPandas()
    expected = pd.DataFrame({"id": [2, 3], "name": ["ALICE", "BOB"]})
    assert_frame_equal(actual, expected)



from __future__ import annotations

import argparse
import contextlib
import time

from pyspark.sql import SparkSession


class TpchBenchmark:
    TABLE_NAMES = ("customer", "lineitem", "nation", "orders", "part", "partsupp", "region", "supplier")

    def __init__(self, url: str, data_path: str, query_path: str):
        self.url = url
        self.data_path = data_path
        self.query_path = query_path

    def _is_remote(self):
        return self.url.startswith("sc://")

    @contextlib.contextmanager
    def spark_session(self):
        builder = SparkSession.builder.remote(self.url) if self._is_remote() else SparkSession.builder.master(self.url)
        spark = builder.appName("TPC-H").getOrCreate()
        for table in self.TABLE_NAMES:
            path = f"{self.data_path}/{table}.parquet"
            df = spark.read.parquet(path)
            df.createOrReplaceTempView(table)
        try:
            yield spark
        finally:
            spark.stop()

    def _run_query(self, spark: SparkSession, query: int, explain: bool):  # noqa: FBT001
        with open(f"{self.query_path}/q{query}.sql") as f:
            for sql in f.read().split(";"):
                sql = sql.strip()  # noqa: PLW2901
                if not sql:
                    continue
                sql = sql.replace("create view", "create temp view")  # noqa: PLW2901
                print(sql)
                if explain:
                    if self._is_remote():
                        df = spark.sql(f"EXPLAIN VERBOSE {sql}")
                        for type_, line in df.collect():
                            if type_:
                                print()
                                print(f"== {type_} ==")
                            print(line)
                    else:
                        df = spark.sql(f"EXPLAIN EXTENDED {sql}")
                        for (line,) in df.collect():
                            print(line)
                else:
                    start_time = time.time()
                    df = spark.sql(sql)
                    rows = df.collect()
                    end_time = time.time()
                    print(f"The query returned {len(rows)} rows and took {end_time - start_time} seconds.")

    def run(self, query: int | None = None, explain: bool = False):  # noqa: FBT001, FBT002
        with self.spark_session() as spark:
            if query is not None:
                self._run_query(spark, query, explain)
            else:
                for query in range(1, 23):
                    self._run_query(spark, query, explain)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--url", type=str, default="sc://localhost:50051")
    parser.add_argument("--data-path", type=str, required=True)
    parser.add_argument("--query-path", type=str, required=True)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--console", action="store_true")
    group.add_argument("--query", type=int, choices=range(1, 23))
    group.add_argument("--query-all", action="store_true")
    group.add_argument("--explain", type=int, choices=range(1, 23))
    args = parser.parse_args()

    benchmark = TpchBenchmark(args.url, args.data_path, args.query_path)
    if args.console:
        with benchmark.spark_session() as spark:
            import code
            import readline
            from rlcompleter import Completer

            namespace = {"spark": spark}
            readline.parse_and_bind("tab: complete")
            readline.set_completer(Completer(namespace).complete)
            code.interact(
                local=namespace,
                banner="Spark TPC-H Data Explorer\nThe Spark session is available as `spark`.",
            )
    elif args.query:
        benchmark.run(args.query)
    elif args.explain:
        benchmark.run(args.explain, explain=True)
    elif args.query_all:
        benchmark.run()


if __name__ == "__main__":
    main()



import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from pyspark.sql import Row
from pyspark.sql.functions import col, lit


def test_dataframe_drop(sail):
    df = sail.createDataFrame([(14, "Tom"), (23, "Alice"), (16, "Bob")], ["age", "name"])
    df2 = sail.createDataFrame([Row(height=80, name="Tom"), Row(height=85, name="Bob")])

    assert_frame_equal(
        df.drop("age").toPandas(),
        pd.DataFrame({"name": ["Tom", "Alice", "Bob"]}),
    )
    assert_frame_equal(
        df.drop(df.age).toPandas(),
        pd.DataFrame({"name": ["Tom", "Alice", "Bob"]}),
    )

    assert_frame_equal(
        df.join(df2, df.name == df2.name, "inner").drop("name").sort("age").toPandas(),
        pd.DataFrame({"age": [14, 16], "height": [80, 85]}),
    )

    df3 = df.join(df2)
    assert_frame_equal(
        df3.select(
            df["age"],
            df["name"].alias("name_left"),
            df2["height"],
            df2["name"].alias("name_right"),
        ).toPandas(),
        pd.DataFrame(
            {
                "age": [14, 23, 16, 14, 23, 16],
                "name_left": ["Tom", "Alice", "Bob", "Tom", "Alice", "Bob"],
                "height": [80, 80, 80, 85, 85, 85],
                "name_right": ["Tom", "Tom", "Tom", "Bob", "Bob", "Bob"],
            }
        ),
    )

    assert_frame_equal(
        df3.drop("name").toPandas(),
        pd.DataFrame({"age": [14, 23, 16, 14, 23, 16], "height": [80, 80, 80, 85, 85, 85]}),
    )

    with pytest.raises(Exception, match="AMBIGUOUS_REFERENCE"):
        df3.drop(col("name")).toPandas()

    df4 = df.withColumn("a.b.c", lit(1))
    assert_frame_equal(
        df4.toPandas(),
        pd.DataFrame({"age": [14, 23, 16], "name": ["Tom", "Alice", "Bob"], "a.b.c": [1, 1, 1]}).astype(
            {"a.b.c": "int32"}
        ),
    )

    assert_frame_equal(
        df4.drop("a.b.c").toPandas(),
        pd.DataFrame({"age": [14, 23, 16], "name": ["Tom", "Alice", "Bob"]}),
    )

    assert_frame_equal(
        df4.drop(col("a.b.c")).toPandas(),
        pd.DataFrame({"age": [14, 23, 16], "name": ["Tom", "Alice", "Bob"], "a.b.c": [1, 1, 1]}).astype(
            {"a.b.c": "int32"}
        ),
    )



import pytest
from pyspark import Row
from pyspark.sql.functions import udf


@pytest.fixture(params=[True, False], ids=["arrow", "non-arrow"])
def arrow(sail, request):
    # The PySpark UDF Arrow mode is buggy since it may produce results
    # different from the non-Arrow mode (e.g. type casting support).
    # Some of the issues are not seen in Sail, so we can write tests and
    # make sure that both modes produce the same results.
    value = "true" if request.param else "false"
    sail.conf.set("spark.sql.execution.pythonUDF.arrow.enabled", value)
    yield
    sail.conf.unset("spark.sql.execution.pythonUDF.arrow.enabled")


@pytest.mark.usefixtures("arrow")
@pytest.mark.skip(reason="not working")
def test_implicit_string_casting(sail):
    # The default UDF return type is "string".
    df = sail.sql("SELECT 1 as a").select(udf(lambda x: x)("a").alias("b"))
    assert df.collect() == [Row(b="1")]

    df = sail.sql("SELECT 1 as a").select(udf(lambda x: [x], returnType="array<string>")("a").alias("b"))
    assert df.collect() == [Row(b=["1"])]


@pytest.mark.usefixtures("arrow")
@pytest.mark.skip(reason="not working")
def test_implicit_binary_casting_invalid_type(sail):
    df = sail.sql("SELECT 1 as a").select(udf(lambda x: x, returnType="binary")("a").alias("b"))
    assert df.collect() == [Row(b=None)]

    df = sail.sql("SELECT 1 as a").select(udf(lambda x: [x], returnType="array<binary>")("a").alias("b"))
    assert df.collect() == [Row(b=[None])]


@pytest.mark.usefixtures("arrow")
@pytest.mark.skip(reason="not working")
def test_implicit_binary_casting_string_type(sail):
    df = sail.sql("SELECT '1' as a").select(udf(lambda x: x, returnType="binary")("a").alias("b"))
    assert df.collect() == [Row(b=bytearray(b"1"))]

    df = sail.sql("SELECT '1' as a").select(udf(lambda x: [x], returnType="array<binary>")("a").alias("b"))
    assert df.collect() == [Row(b=[bytearray(b"1")])]



# Configuration file for the Sphinx documentation builder.
from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import quote

from sphinxcontrib.serializinghtml import JSONHTMLBuilder

if TYPE_CHECKING:
    from sphinx.application import Sphinx

project = "Sail"
copyright = "LakeSail, Inc."  # noqa: A001
author = "LakeSail"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

autosummary_generate = True


class _JSONHTMLBuilder(JSONHTMLBuilder):
    def get_target_uri(self, docname: str, typ: str | None = None) -> str:  # noqa: ARG002
        # We generate the URI in the same way as the HTML builder.
        return quote(docname) + self.link_suffix


def setup(app: Sphinx):
    app.add_builder(_JSONHTMLBuilder, override=True)



import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from pyspark import Row


@pytest.fixture(scope="module", autouse=True)
def customer(sail):
    sail.createDataFrame(
        [
            Row(id=1, name="Alice", age=34),
            Row(id=2, name="Bob", age=36),
            Row(id=3, name="Charlie", age=30),
        ]
    ).createOrReplaceTempView("customer")
    yield
    sail.catalog.dropTempView("customer")


@pytest.fixture(scope="module", autouse=True)
def order(sail):
    sail.createDataFrame(
        [
            Row(id=1, customer_id=1, amount=100),
            Row(id=2, customer_id=2, amount=200),
            Row(id=3, customer_id=2, amount=300),
        ]
    ).createOrReplaceTempView("order")
    yield
    sail.catalog.dropTempView("order")


def test_join_group_by(sail):
    # This example is inspired by the Ballista documentation:
    #   https://datafusion.apache.org/ballista/contributors-guide/architecture.html
    actual = sail.sql(
        """
        SELECT customer.name AS name, SUM(order.amount) AS total
        FROM customer
        JOIN order ON customer.id = order.customer_id
        GROUP BY customer.name
        ORDER BY total DESC
        """
    ).toPandas()
    expected = pd.DataFrame({"name": ["Bob", "Alice"], "total": [500, 100]})
    assert_frame_equal(actual, expected)



import code
import platform
import readline
from rlcompleter import Completer

import pyspark
from pyspark.sql import SparkSession


def run_pyspark_shell(port: int):
    spark = SparkSession.builder.remote(f"sc://localhost:{port}").getOrCreate()
    namespace = {"spark": spark}
    readline.parse_and_bind("tab: complete")
    readline.set_completer(Completer(namespace).complete)

    python_version = platform.python_version()
    (build_number, build_date) = platform.python_build()
    # Simulate the messages similar to `shell.py` in PySpark.
    banner = rf"""Welcome to
      ____              __
     / __/__  ___ _____/ /__
    _\ \/ _ \/ _ `/ __/  '_/
   /__ / .__/\_,_/_/ /_/\_\   version {pyspark.__version__}
      /_/

Using Python version {python_version} ({build_number}, {build_date})
Client connected to the Sail Spark Connect server at localhost:{port}
SparkSession available as 'spark'."""
    code.interact(local=namespace, banner=banner, exitmsg="")



from __future__ import annotations

import ctypes
import decimal
import itertools
from typing import Any, Callable, Iterator, Sequence, Union

import pandas as pd
import pyarrow as pa
from pyspark.sql.pandas.serializers import ArrowStreamPandasUDFSerializer, ArrowStreamPandasUDTFSerializer
from pyspark.sql.pandas.types import from_arrow_type
from pyspark.sql.types import Row

_PYARROW_HAS_VIEW_TYPES = all(hasattr(pa, x) for x in ("list_view", "large_list_view", "string_view", "binary_view"))

if _PYARROW_HAS_VIEW_TYPES:
    _PYARROW_LIST_TYPES = (pa.ListType, pa.LargeListType, pa.FixedSizeListType, pa.ListViewType, pa.LargeListViewType)
    _PYARROW_LIST_ARRAY_TYPES = (
        pa.ListArray,
        pa.LargeListArray,
        pa.FixedSizeListArray,
        pa.ListViewArray,
        pa.LargeListViewArray,
    )
    _PyArrowListType = Union[pa.ListType, pa.LargeListType, pa.FixedSizeListType, pa.ListViewType, pa.LargeListViewType]

    def _pyarrow_is_string(t: pa.DataType) -> bool:
        return pa.types.is_string(t) or pa.types.is_large_string(t) or pa.types.is_string_view(t)

    def _pyarrow_is_binary(t: pa.DataType) -> bool:
        return pa.types.is_binary(t) or pa.types.is_large_binary(t) or pa.types.is_binary_view(t)
else:
    _PYARROW_LIST_TYPES = (pa.ListType, pa.LargeListType, pa.FixedSizeListType)
    _PYARROW_LIST_ARRAY_TYPES = (pa.ListArray, pa.LargeListArray, pa.FixedSizeListArray)
    _PyArrowListType = Union[pa.ListType, pa.LargeListType, pa.FixedSizeListType]

    def _pyarrow_is_string(t: pa.DataType) -> bool:
        return pa.types.is_string(t) or pa.types.is_large_string(t)

    def _pyarrow_is_binary(t: pa.DataType) -> bool:
        return pa.types.is_binary(t) or pa.types.is_large_binary(t)


try:
    from itertools import batched
except ImportError:

    def batched(iterable, n):
        it = iter(iterable)
        while chunk := tuple(itertools.islice(it, n)):
            yield chunk


class Converter:
    """
    A converter that converts between PySpark data and Arrow data.
    When matching PySpark data to the Arrow data type, invalid data is converted to null.
    This is the similar behavior in the Scala implementation of PySpark UDF [1].

    * [1] `org.apache.spark.sql.execution.python.EvaluatePython#makeFromJava`
    """

    def __init__(self, data_type: pa.DataType):
        self._data_type = data_type

    def to_pyspark(self, data: pa.Array) -> Sequence[Any]:
        raise NotImplementedError

    def from_pyspark(self, data: Sequence[Any]) -> pa.Array:
        raise NotImplementedError


def _get_converter(t: pa.DataType) -> Converter:
    if pa.types.is_null(t):
        return NullConverter(t)
    if pa.types.is_boolean(t):
        return BooleanConverter(t)
    if pa.types.is_int8(t):
        return IntegerConverter(t, ctypes.c_int8)
    if pa.types.is_int16(t):
        return IntegerConverter(t, ctypes.c_int16)
    if pa.types.is_int32(t):
        return IntegerConverter(t, ctypes.c_int32)
    if pa.types.is_int64(t):
        return IntegerConverter(t, ctypes.c_int64)
    if pa.types.is_uint8(t):
        return IntegerConverter(t, ctypes.c_uint8)
    if pa.types.is_uint16(t):
        return IntegerConverter(t, ctypes.c_uint16)
    if pa.types.is_uint32(t):
        return IntegerConverter(t, ctypes.c_uint32)
    if pa.types.is_uint64(t):
        return IntegerConverter(t, ctypes.c_uint64)
    if pa.types.is_floating(t):
        return FloatConverter(t)
    if pa.types.is_decimal(t):
        return DecimalConverter(t)
    if pa.types.is_time(t) or pa.types.is_timestamp(t) or pa.types.is_date(t) or pa.types.is_duration(t):
        return DateTimeConverter(t)
    if _pyarrow_is_string(t):
        return StringConverter(t)
    if _pyarrow_is_binary(t):
        return BinaryConverter(t)
    if isinstance(t, _PYARROW_LIST_TYPES):
        return ArrayConverter(t)
    if isinstance(t, pa.MapType):
        return MapConverter(t)
    if isinstance(t, pa.StructType):
        return StructConverter(t)
    msg = f"unsupported data type: {t}"
    raise ValueError(msg)


def _raise_for_row(data: Any):
    if isinstance(data, Row):
        # Simulate the exception when the JVM receives an invalid row for the data type.
        msg = "net.razorvine.pickle.PickleException: expected zero arguments for construction of ClassDict (for pyspark.sql.types._create_row)."
        raise TypeError(msg)


class ScalarConverter(Converter):
    def to_pyspark(self, data: pa.Array) -> Sequence[Any]:
        return [self._to_pyspark_value(x) for x in data.to_pylist()]

    def from_pyspark(self, data: Sequence[Any]) -> pa.Array:
        return pa.array([self._from_pyspark_value(x) for x in data], type=self._data_type)

    def _to_pyspark_value(self, data: Any) -> Any:
        raise NotImplementedError

    def _from_pyspark_value(self, data: Any) -> Any:
        raise NotImplementedError


class PrimitiveConverter(ScalarConverter):
    def __init__(self, data_type: pa.DataType):
        super().__init__(data_type)
        self._spark_data_type = from_arrow_type(data_type)

    def _to_pyspark_value(self, data: Any) -> Any:
        # Reference: `pyspark.sql.types._create_row_inbound_converter`
        return self._spark_data_type.fromInternal(data)

    def _from_pyspark_value(self, data: Any) -> Any:
        return self._from_python_primitive(self._spark_data_type.toInternal(data))

    def _from_python_primitive(self, data: Any) -> Any:
        raise NotImplementedError


class NullConverter(PrimitiveConverter):
    def _from_python_primitive(self, data: Any) -> Any:
        _raise_for_row(data)
        return None


class BooleanConverter(PrimitiveConverter):
    def _from_python_primitive(self, data: Any) -> Any:
        _raise_for_row(data)
        if isinstance(data, bool):
            return data
        return None


class IntegerConverter(PrimitiveConverter):
    def __init__(self, data_type: pa.DataType, ctype: Any):
        super().__init__(data_type)
        self._ctype = ctype

    def _from_python_primitive(self, data: Any) -> Any:
        _raise_for_row(data)
        if isinstance(data, int):
            return self._ctype(data).value
        return None


class FloatConverter(PrimitiveConverter):
    def _from_python_primitive(self, data: Any) -> Any:
        _raise_for_row(data)
        if isinstance(data, float):
            return data
        return None


class DecimalConverter(PrimitiveConverter):
    def _from_python_primitive(self, data: Any) -> Any:
        _raise_for_row(data)
        if isinstance(data, decimal.Decimal):
            return data
        return None


class DateTimeConverter(ScalarConverter):
    # No conversion is performed since pyarrow already handles conversion for
    # types in the `datetime` module.

    def _to_pyspark_value(self, data: Any) -> Any:
        return data

    def _from_pyspark_value(self, data: Any) -> Any:
        return data


def _to_string(data: Any) -> str | None:
    """Converts data to string where the behavior is similar to the Java `Object.toString()` method."""
    _raise_for_row(data)
    if data is None:
        return None
    if isinstance(data, str):
        return data
    if isinstance(data, bool):
        return "true" if data else "false"
    if isinstance(data, (list, tuple)):
        items = ", ".join(_to_string(x) for x in data)
        return f"[{items}]"
    if isinstance(data, dict):
        items = ", ".join(f"{_to_string(k)}={_to_string(v)}" for k, v in data.items())
        return f"{{{items}}}"
    return str(data)


class StringConverter(ScalarConverter):
    def _to_pyspark_value(self, data: Any) -> Any:
        return data

    def _from_pyspark_value(self, data: Any) -> Any:
        return _to_string(data)


def _to_bytes(data: Any) -> bytes | None:
    _raise_for_row(data)
    if data is None:
        return None
    if isinstance(data, str):
        return data.encode("utf-8")
    if isinstance(data, bytes):
        return data
    if isinstance(data, bytearray):
        return bytes(data)
    return None


class BinaryConverter(ScalarConverter):
    def _to_pyspark_value(self, data: Any) -> Any:
        return None if data is None else bytearray(data)

    def _from_pyspark_value(self, data: Any) -> Any:
        return _to_bytes(data)


class ArrayConverter(Converter):
    def __init__(self, data_type: _PyArrowListType):
        super().__init__(data_type)
        self._value_converter = _get_converter(data_type.value_type)

    def to_pyspark(self, array: pa.Array) -> Sequence[Any]:
        if not isinstance(array, _PYARROW_LIST_ARRAY_TYPES):
            msg = f"invalid data type for array: {type(array)}"
            raise TypeError(msg)
        values = self._value_converter.to_pyspark(array.flatten())
        offsets = array.offsets.to_pylist()
        valid = array.is_valid().to_pylist()
        result = []
        for i in range(len(array)):
            if not valid[i]:
                result.append(None)
            else:
                (start, end) = (offsets[i], offsets[i + 1])
                result.append(values[start:end])
        return result

    def from_pyspark(self, data: Sequence[Any]) -> pa.Array:
        (values, offsets) = [], []
        end = 0
        for x in data:
            _raise_for_row(x)
            if x is None or not isinstance(x, (list, tuple)):
                offsets.append(None)
            else:
                offsets.append(end)
                values.extend(x)
                end += len(x)
        offsets.append(end)
        return pa.ListArray.from_arrays(pa.array(offsets, type=pa.int32()), self._value_converter.from_pyspark(values))


class MapConverter(Converter):
    def __init__(self, data_type: pa.MapType):
        super().__init__(data_type)
        self._key_converter = _get_converter(data_type.key_type)
        self._value_converter = _get_converter(data_type.item_type)

    def to_pyspark(self, array: pa.Array) -> Sequence[Any]:
        if not isinstance(array, pa.MapArray):
            msg = f"invalid data type for map: {type(array)}"
            raise TypeError(msg)
        keys = self._key_converter.to_pyspark(array.keys)
        values = self._value_converter.to_pyspark(array.items)
        offsets = array.offsets.to_pylist()
        valid = array.is_valid().to_pylist()
        result = []
        for i in range(len(array)):
            if not valid[i]:
                result.append(None)
            else:
                (start, end) = (offsets[i], offsets[i + 1])
                result.append(dict(zip(keys[start:end], values[start:end])))
        return result

    def from_pyspark(self, data: Sequence[Any]) -> pa.Array:
        (keys, values, offsets) = [], [], []
        end = 0
        for x in data:
            _raise_for_row(x)
            if x is None or not isinstance(x, dict):
                offsets.append(None)
            else:
                offsets.append(end)
                keys.extend(x.keys())
                values.extend(x.values())
                end += len(x)
        offsets.append(end)
        return pa.MapArray.from_arrays(
            pa.array(offsets, type=pa.int32()),
            self._key_converter.from_pyspark(keys),
            self._value_converter.from_pyspark(values),
        )


class StructConverter(Converter):
    def __init__(self, data_type: pa.StructType):
        super().__init__(data_type)
        try:
            self._fields = data_type.fields
        except AttributeError:
            self._fields = [data_type.field(i) for i in range(data_type.num_fields)]
        self._field_converters = [_get_converter(f.type) for f in self._fields]
        self._spark_data_type = from_arrow_type(data_type)

    def to_pyspark(self, array: pa.Array) -> Sequence[Any]:
        if not isinstance(array, pa.StructArray):
            msg = f"invalid data type for struct: {type(array)}"
            raise TypeError(msg)
        columns = [c.to_pyspark(col) for col, c in zip(array.flatten(), self._field_converters)]
        return [self._spark_data_type.fromInternal(x) for x in zip(*columns)]

    def from_pyspark(self, data: Sequence[Any]) -> pa.Array:
        n = len(self._fields)
        columns = [[] for _ in range(n)]
        mask = []
        for x in data:
            if x is None:
                mask.append(True)
                for i in range(n):
                    columns[i].append(None)
            else:
                mask.append(False)
                for i, v in enumerate(self._spark_data_type.toInternal(x)):
                    columns[i].append(v)
        return pa.StructArray.from_arrays(
            [c.from_pyspark(col) for col, c in zip(columns, self._field_converters)],
            fields=self._fields,
            mask=pa.array(mask, type=pa.bool_()),
        )


def _pandas_to_arrow_array(data, data_type: pa.DataType, serializer: ArrowStreamPandasUDFSerializer) -> pa.Array:
    if serializer._struct_in_pandas == "dict" and pa.types.is_struct(data_type):  # noqa: SLF001
        return serializer._create_struct_array(data, data_type)  # noqa: SLF001
    return serializer._create_array(data, data_type, arrow_cast=serializer._arrow_cast)  # noqa: SLF001


def _named_arrays_to_pandas(
    data: Sequence[pa.Array], names: Sequence[str], serializer: ArrowStreamPandasUDFSerializer
) -> Sequence[pd.Series]:
    inputs = [serializer.arrow_to_pandas(x) for x in data]
    for x, name in zip(inputs, names):
        x.name = name
    return inputs


class PySparkBatchUdf:
    def __init__(self, udf: Callable[..., Any], input_types: Sequence[pa.DataType], output_type: pa.DataType):
        self._udf = udf
        self._input_converters = [_get_converter(t) for t in input_types]
        self._output_converter = _get_converter(output_type)

    def __call__(self, args: list[pa.Array], num_rows: int) -> pa.Array:
        if len(args) > 0:
            inputs = [c.to_pyspark(a) for a, c in zip(args, self._input_converters)]
            output = list(self._udf(None, zip(*inputs)))
        else:
            output = list(self._udf(None, itertools.repeat((), num_rows)))
        return self._output_converter.from_pyspark(output)


class PySparkArrowBatchUdf:
    def __init__(self, udf: Callable[..., Any], config):
        self._udf = udf
        self._serializer = ArrowStreamPandasUDFSerializer(
            timezone=config.session_timezone,
            safecheck=config.arrow_convert_safely,
            assign_cols_by_name=config.assign_columns_by_name,
            df_for_struct=False,
            struct_in_pandas="row",
            ndarray_as_list=True,
            arrow_cast=True,
        )

    def __call__(self, args: list[pa.Array], _num_rows: int) -> pa.Array:
        inputs = tuple(self._serializer.arrow_to_pandas(a) for a in args)
        [(output, output_type)] = list(self._udf(None, (inputs,)))
        return _pandas_to_arrow_array(output, output_type, self._serializer)


class PySparkScalarPandasUdf:
    def __init__(
        self,
        udf: Callable[..., Any],
        config,
    ):
        self._udf = udf
        self._serializer = ArrowStreamPandasUDFSerializer(
            timezone=config.session_timezone,
            safecheck=config.arrow_convert_safely,
            assign_cols_by_name=config.assign_columns_by_name,
            df_for_struct=True,
            struct_in_pandas="dict",
            ndarray_as_list=False,
            arrow_cast=False,
        )

    def __call__(self, args: list[pa.Array], _num_rows: int) -> pa.Array:
        inputs = tuple(self._serializer.arrow_to_pandas(x) for x in args)
        [(output, output_type)] = list(self._udf(None, (inputs,)))
        return _pandas_to_arrow_array(output, output_type, self._serializer)


class PySparkScalarPandasIterUdf:
    def __init__(
        self,
        udf: Callable[..., Any],
        config,
    ):
        self._udf = udf
        self._serializer = ArrowStreamPandasUDFSerializer(
            timezone=config.session_timezone,
            safecheck=config.arrow_convert_safely,
            assign_cols_by_name=config.assign_columns_by_name,
            df_for_struct=True,
            struct_in_pandas="dict",
            ndarray_as_list=False,
            arrow_cast=False,
        )

    def __call__(self, args: list[pa.Array], _num_rows: int) -> pa.Array:
        inputs = tuple(self._serializer.arrow_to_pandas(x) for x in args)
        [(output, output_type)] = list(self._udf(None, [inputs]))
        return _pandas_to_arrow_array(output, output_type, self._serializer)


class PySparkGroupAggUdf:
    def __init__(
        self,
        udf: Callable[..., Any],
        input_names: Sequence[str],
        config,
    ):
        self._udf = udf
        self._input_names = input_names
        self._serializer = ArrowStreamPandasUDFSerializer(
            timezone=config.session_timezone,
            safecheck=config.arrow_convert_safely,
            assign_cols_by_name=config.assign_columns_by_name,
            df_for_struct=True,
            struct_in_pandas="dict",
            ndarray_as_list=False,
            arrow_cast=False,
        )

    def __call__(self, args: list[pa.Array]) -> pa.Array:
        inputs = _named_arrays_to_pandas(args, self._input_names, self._serializer)
        [(output, output_type)] = list(self._udf(None, (inputs,)))
        return _pandas_to_arrow_array(output, output_type, self._serializer)


class PySparkGroupMapUdf:
    def __init__(
        self,
        udf: Callable[..., Any],
        input_names: Sequence[str],
        config,
    ):
        self._udf = udf
        self._input_names = input_names
        self._serializer = ArrowStreamPandasUDFSerializer(
            timezone=config.session_timezone,
            safecheck=config.arrow_convert_safely,
            assign_cols_by_name=config.assign_columns_by_name,
            df_for_struct=True,
            struct_in_pandas="dict",
            ndarray_as_list=False,
            arrow_cast=False,
        )

    def __call__(self, args: list[pa.Array]) -> pa.Array:
        inputs = _named_arrays_to_pandas(args, self._input_names, self._serializer)
        [[(output, output_type)]] = list(self._udf(None, (inputs,)))
        return _pandas_to_arrow_array(output, output_type, self._serializer)


class PySparkCoGroupMapUdf:
    def __init__(
        self,
        udf: Callable[..., Any],
        left_names: Sequence[str],
        right_names: Sequence[str],
        config,
    ):
        self._udf = udf
        self._left_names = left_names
        self._right_names = right_names
        self._serializer = ArrowStreamPandasUDFSerializer(
            timezone=config.session_timezone,
            safecheck=config.arrow_convert_safely,
            assign_cols_by_name=config.assign_columns_by_name,
            df_for_struct=True,
            struct_in_pandas="dict",
            ndarray_as_list=False,
            arrow_cast=False,
        )

    def __call__(self, left: list[pa.Array], right: list[pa.Array]) -> pa.Array:
        args = [
            _named_arrays_to_pandas(left, self._left_names, self._serializer),
            _named_arrays_to_pandas(right, self._right_names, self._serializer),
        ]
        [[(output, output_type)]] = list(self._udf(None, (args,)))
        return _pandas_to_arrow_array(output, output_type, self._serializer)


class PySparkMapPandasIterUdf:
    def __init__(
        self,
        udf: Callable[..., Iterator[pd.DataFrame]],
        config,
    ):
        self._udf = udf
        self._serializer = ArrowStreamPandasUDFSerializer(
            timezone=config.session_timezone,
            safecheck=config.arrow_convert_safely,
            assign_cols_by_name=config.assign_columns_by_name,
            df_for_struct=True,
            struct_in_pandas="dict",
            ndarray_as_list=False,
            arrow_cast=False,
        )

    def __call__(self, args: Iterator[pa.RecordBatch]) -> Iterator[pa.RecordBatch]:
        output = self._udf(None, ((self._convert_input(x),) for x in args))
        return (self._convert_output(x, t) for x, t in output)

    def _convert_input(self, batch: pa.RecordBatch) -> pd.DataFrame:
        return self._serializer.arrow_to_pandas(batch.to_struct_array())

    def _convert_output(self, df: pd.DataFrame, data_type: pa.DataType) -> pa.RecordBatch:
        array = _pandas_to_arrow_array(df, data_type, self._serializer)
        return pa.RecordBatch.from_struct_array(array)


class PySparkMapArrowIterUdf:
    def __init__(self, udf: Callable[..., Iterator[pa.RecordBatch]]):
        self._udf = udf

    def __call__(self, args: Iterator[pa.RecordBatch]) -> Iterator[pa.RecordBatch]:
        output = self._udf(None, ((x,) for x in args))
        return (x for x, _ in output)


class PySparkTableUdf:
    def __init__(
        self,
        udf: Callable[..., Any],
        input_types: Sequence[pa.DataType],
        passthrough_columns: int,
        output_schema: pa.Schema,
        config,
    ):
        self._udf = udf
        self._passthrough_columns = passthrough_columns
        self._batch_size = config.arrow_max_records_per_batch
        if self._batch_size <= 0:
            msg = f"invalid batch size: {self._batch_size}"
            raise ValueError(msg)
        self._output_schema = output_schema
        self._input_converters = [_get_converter(t) for t in input_types]
        self._output_converter = StructConverter(
            pa.struct([output_schema.field(i) for i in range(len(output_schema.names))])
        )

    def __call__(self, args: Iterator[pa.RecordBatch]) -> Iterator[pa.RecordBatch]:
        for output in batched(self._iter_output_rows(args), self._batch_size):
            yield pa.RecordBatch.from_struct_array(self._output_converter.from_pyspark(output))

    def _iter_input_rows(self, args: Iterator[pa.RecordBatch]) -> Iterator[tuple]:
        for batch in args:
            arrays = batch.to_struct_array().flatten()
            if len(arrays) > 0:
                inputs = tuple(c.to_pyspark(a) for a, c in zip(arrays, self._input_converters))
                yield from zip(*inputs)
            else:
                yield ()

    def _iter_output_rows(self, args: Iterator[pa.RecordBatch]) -> Iterator[list]:
        (rows1, rows2) = itertools.tee(self._iter_input_rows(args))
        inputs = (x[self._passthrough_columns :] for x in rows2)
        outputs = self._udf(None, inputs)
        empty = tuple([None] * len(self._output_schema.names))
        last = tuple([None] * len(self._input_converters))
        for row, it in itertools.zip_longest(rows1, outputs):
            if row is None:
                row = last  # noqa: PLW2901
            passthrough = row[: self._passthrough_columns]
            for out in it:
                if out is None:
                    yield passthrough + empty
                else:
                    yield passthrough + tuple(out)
            last = row


class PySparkArrowTableUdf:
    def __init__(
        self,
        udf: Callable[..., Iterator[pa.RecordBatch]],
        input_names: Sequence[str],
        passthrough_columns: int,
        output_schema: pa.Schema,
        config,
    ):
        self._udf = udf
        self._input_names = input_names
        self._passthrough_columns = passthrough_columns
        self._output_schema = output_schema
        self._output_type = pa.struct([output_schema.field(i) for i in range(len(output_schema.names))])
        self._serializer = ArrowStreamPandasUDTFSerializer(
            timezone=config.session_timezone, safecheck=config.arrow_convert_safely
        )

    def __call__(self, args: Iterator[pa.RecordBatch]) -> Iterator[pa.RecordBatch]:
        for output in self._iter_output(args):
            array = self._serializer._create_struct_array(output, self._output_type)  # noqa: SLF001
            yield pa.RecordBatch.from_struct_array(array)

    def _iter_input(self, args: Iterator[pa.RecordBatch]) -> Iterator[tuple[pd.Series]]:
        for batch in args:
            arrays = batch.to_struct_array().flatten()
            yield tuple(self._serializer.arrow_to_pandas(x) for x in arrays)

    def _iter_output(self, args: Iterator[pa.RecordBatch]) -> Iterator[pd.DataFrame]:
        (batches1, batches2) = itertools.tee(self._iter_input(args))
        inputs = (x[self._passthrough_columns :] for x in batches2)
        outputs = self._udf(None, inputs)
        last = None
        for passthrough, (out, _) in itertools.zip_longest(self._iter_passthrough(batches1), outputs):
            if out is None or len(out) == 0:
                continue
            df = pd.DataFrame(index=out.index)
            if passthrough is None:
                passthrough = last  # noqa: PLW2901
            if passthrough is not None:
                for v, name in zip(passthrough, self._input_names):
                    df[name] = [v] * len(out)
            else:
                for name in self._input_names:
                    df[name] = [None] * len(out)
            for col in out:
                df[col] = out[col]
            yield df
            last = passthrough

    def _iter_passthrough(self, batches: Iterator[tuple[pd.Series]]) -> Iterator[tuple]:
        if self._passthrough_columns > 0:
            for batch in batches:
                yield from zip(*batch[: self._passthrough_columns])
        else:
            for batch in batches:
                if len(batch) > 0:
                    first, *_ = batch
                    for _ in range(len(first)):
                        yield ()
                else:
                    yield ()



import pysail
from pysail import _native


def test_version():
    v1, v2, v3, *_ = pysail.__version__.split(".")
    major, minor, patch = _native._SAIL_VERSION.split(".")  # noqa: SLF001
    assert v1 == major
    assert v2 == minor
    assert v3 == patch or v3.startswith((f"{patch}a", f"{patch}b", f"{patch}rc"))



import pandas as pd
from pandas.testing import assert_frame_equal


def test_sort(sail):
    # Reference: pyspark.sql.tests.connect.test_connect_basic.SparkConnectBasicTests.test_sort
    # We need to make sure sorting works since we have patched PySpark tests to ignore row order.
    query = """
        SELECT * FROM VALUES
        (false, 1, NULL), (false, NULL, 2.0), (NULL, 3, 3.0)
        AS tab(a, b, c)
    """
    # +-----+----+----+
    # |    a|   b|   c|
    # +-----+----+----+
    # |false|   1|NULL|
    # |false|NULL| 2.0|
    # | NULL|   3| 3.0|
    # +-----+----+----+

    df = sail.sql(query)
    assert_frame_equal(
        df.sort("a").toPandas()[["a"]],
        pd.DataFrame({"a": [None, False, False]}),
    )
    assert_frame_equal(
        df.sort("a").toPandas().sort_values(by=["a", "b"], ignore_index=True),
        pd.DataFrame(
            {"a": [False, False, None], "b": [1, None, 3], "c": [None, 2.0, 3.0]},
        ).astype({"c": object}),
    )
    assert_frame_equal(
        df.sort("c").toPandas(),
        pd.DataFrame(
            {"a": [False, False, None], "b": [1, None, 3], "c": [None, 2.0, 3.0]},
        ).astype({"c": object}),
    )
    assert_frame_equal(
        df.sort("b").toPandas(),
        pd.DataFrame(
            {"a": [False, False, None], "b": [None, 1, 3], "c": [2.0, None, 3.0]},
        ).astype({"c": object}),
    )
    assert_frame_equal(
        df.sort(df.c, "b").toPandas(),
        pd.DataFrame(
            {"a": [False, False, None], "b": [1, None, 3], "c": [None, 2.0, 3.0]},
        ).astype({"c": object}),
    )
    assert_frame_equal(
        df.sort(df.c.desc(), "b").toPandas(),
        pd.DataFrame(
            {"a": [None, False, False], "b": [3, None, 1], "c": [3.0, 2.0, None]},
        ).astype({"c": object}),
    )
    assert_frame_equal(
        df.sort(df.c.desc(), df.a.asc()).toPandas(),
        pd.DataFrame(
            {"a": [None, False, False], "b": [3, None, 1], "c": [3.0, 2.0, None]},
        ).astype({"c": object}),
    )
